/*
 * Copyright (C) 2014 digitalSTROM.org, Schlieren, Switzerland
 *
 * Authors: Andreas Fenkart, <andreas.fenkart@dev.digitalstrom.org>
 *
 * digitalSTROM ds485-base library is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * digitalSTROM ds485-base library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with digitalSTROM ds485-base library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "socket_send.h"

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>

#include "log.h"

#if defined(__APPLE__) || defined(__MACH__)
# ifndef MSG_NOSIGNAL
#   define MSG_NOSIGNAL SO_NOSIGPIPE
# endif
#endif

int socket_send(int socket, const uint8_t* data, size_t dlen)
{
    struct timespec tim = { 0, 50000000 }; /* 50 ms */
    ssize_t nbytes, todo = dlen;
    int ret, retries = 0;

    while (retries < 3 && todo > 0) {
        nbytes = send(socket, data, todo, MSG_NOSIGNAL);
        if (nbytes == -1) {
            ret = errno;
            if (ret == EINTR) {
                /* interrupted syscall */
                continue;
            } else if (ret == EAGAIN || ret == EWOULDBLOCK) {
                retries++;
                nanosleep(&tim , NULL);
                continue;
            }
            return -ret; /* peer has disconnected */
        }
        data += nbytes;
        todo -= nbytes;
    }

    if (todo > 0) {
        return -ETIMEDOUT;
    }
    return 0;
}

/**
 * socket_sendv -- send multiple buffer in one message
 *
 * google keywords: nagle writev tcp_nodelay tcp_cork
 * http://baus.net/on-tcp_cork/
 *
 * CAUTION: the kernel will breakup the message into several
 * socket writes, if necessary. Which could introduce a
 * considerable latency.
 */
int socket_sendv(int socket, struct iovec *iov, int iovcnt)
{
    struct timespec tim = { 0, 50000000 }; /* 50 ms */
    struct msghdr mh;
    ssize_t total = 0, ret;
    int i;

    /*
     * the only data transferred with the syscall is the
     * iovec, the data itself is read from within the kernel
     * accessing user space directly
     * the kernel will iterate over the iovec, breaking
     * it up into several writes, hence the sendmsg will
     * always write the full message or fail, no partial
     * writes possible
     *
     * http://lxr.free-electrons.com/source/fs/read_write.c#L653
     * http://lxr.free-electrons.com/source/net/socket.c#L636
     * see __sys_sendmsg, copy_msghdr_from_user, import_iovec
     */

    for (i = 0; i < iovcnt; i++) {
        total += iov[i].iov_len;
    }

    /*
     * That said of course there are limits, the size of
     * iovec is constrained to UIOV_MAXIOV, typically 1024,
     * and the messages size is truncated to
     * MAX_RW_COUNT(~0L >> 1).
     * We will not implement retry mechanim because these
     * limits are much bigger than any message we ever
     * transfer, instead we return with an error or sendmsg
     * will
     *
     * http://lxr.free-electrons.com/source/fs/read_write.c#L64
     * see rw_copy_check_uvector
     */
    if (total > INT_MAX - 4095) {
        return -EMSGSIZE;
    }

    memset(&mh, 0, sizeof(mh));
    mh.msg_iovlen = iovcnt;
    mh.msg_iov = iov;

    for (i = 0; i < 3; i++) {
        ret = sendmsg(socket, &mh, MSG_NOSIGNAL);
        if (ret == -1) {
            ret = errno;
            if (ret == EINTR) {
                /* interrupted syscall */
                continue;
            } else if (ret == EAGAIN || ret == EWOULDBLOCK) {
                nanosleep(&tim , NULL);
                continue;
            }

            /* see man write -- other errors are unrecoverable */
            return -ret;
        }

        if (ret < total) {
            /*
             * As described above, the kernel is accessing user space
             * memory directly, iovec only declares its location. Hence it
             * can split it into packets at will.
             * The only reason left, to explain a short write, seems a
             * client side issued disconnect in the middle of a
             * transaction
             * http://redmine.digitalstrom.org/issues/9038
             */
            report(LOG_NOTICE, "%s short write written(%d) < total(%d)\n",
                   __func__, ret, total);
            return -ECONNABORTED;
        }

        assert(ret == total);
        return 0;
    }

    return -ETIMEDOUT;
}
